@@ -37,6 +37,7 @@ use File::Temp qw/ tempfile tempdir /;
 use Cwd qw(abs_path getcwd);
 use autouse 'Carp' => qw(longmess confess);
 use Scalar::Util 'reftype';
+use FindBin;
 
 our @ISA = qw(Exporter);
 #
@@ -52,7 +53,7 @@ our @EXPORT = qw(
 		 progress_message3
                 );
 
-our @EXPORT_OK = qw( $shorewall_dir initialize read_a_line1 set_config_path shorewall);
+our @EXPORT_OK = qw( $shorewall_dir initialize set_config_path shorewall);
 
 our %EXPORT_TAGS = ( internal => [ qw( create_temp_script
 				       finalize_script
@@ -96,11 +97,15 @@ our %EXPORT_TAGS = ( internal => [ qw( create_temp_script
 				       close_file
 				       push_open
 				       pop_open
+				       push_params
+				       pop_params
 				       read_a_line
 				       validate_level
 				       which
 				       qt
 				       ensure_config_path
+				       add_param
+				       export_params
 				       get_configuration
 				       require_capability
 				       have_capability
@@ -122,7 +127,7 @@ our %EXPORT_TAGS = ( internal => [ qw( create_temp_script
 				       $debug
 				       %config
 				       %globals
-				       %params
+				       %config_files
 
 		                       F_IPV4
 		                       F_IPV6
@@ -133,7 +138,7 @@ our %EXPORT_TAGS = ( internal => [ qw( create_temp_script
 
 Exporter::export_ok_tags('internal');
 
-our $VERSION = '4.4_15';
+our $VERSION = '4.4_19';
 
 #
 # describe the current command, it's present progressive, and it's completion.
@@ -255,9 +260,54 @@ our %capdesc = ( NAT_ENABLED     => 'NAT',
 		 FWMARK_RT_MASK  => 'fwmark route mask',
 		 MARK_ANYWHERE   => 'Mark in any table',
 		 HEADER_MATCH    => 'Header Match',
+		 ACCOUNT_TARGET  => 'ACCOUNT Target',
 		 CAPVERSION      => 'Capability Version',
 		 KERNELVERSION   => 'Kernel Version',
 	       );
+
+our %config_files = ( #accounting      => 1,
+		      actions          => 1,
+		      blacklist        => 1,
+		      clear            => 1,
+		      ecn              => 1,
+		      findgw           => 1,
+		      hosts            => 1,
+		      init             => 1,
+		      initdone         => 1,
+		      interfaces       => 1,
+		      isusable         => 1,
+		      maclist          => 1,
+		      masq             => 1,
+		      nat              => 1,
+		      netmap           => 1,
+		      notrack          => 1,
+		      params           => 1,
+		      policy           => 1,
+		      providers        => 1,
+		      proxyarp         => 1,
+		      refresh          => 1,
+		      refreshed        => 1,
+		      restored         => 1,
+		      route_rules      => 1,
+		      routes           => 1,
+		      routestopped     => 1,
+		      rules            => 1,
+		      scfilter         => 1,
+		      secmarks         => 1,
+		      start            => 1,
+		      started          => 1,
+		      stop             => 1,
+		      stopped          => 1,
+		      tcclasses        => 1,
+		      tcclear          => 1,
+		      tcdevices        => 1,
+		      tcfilters        => 1,
+		      tcinterfaces     => 1,
+		      tcpri            => 1,
+		      tcrules          => 1,
+		      tos              => 1,
+		      tunnels          => 1,
+		      zones            => 1 );
 #
 # Directories to search for configuration files
 #
@@ -274,6 +324,14 @@ our @openstack;
 # From the params file
 #
 our %params;
+#
+# Entries that the compiler adds to %params
+#
+our %compiler_params;
+#
+# Action parameters
+#
+our %actparms;
 
 our $currentline;             # Current config file line image
 our $currentfile;             # File handle reference
@@ -304,7 +362,7 @@ use constant { MIN_VERBOSITY => -1,
 	       MAX_VERBOSITY => 2 ,
 	       F_IPV4 => 4,
 	       F_IPV6 => 6,
-	   };
+	     };
 
 our %validlevels;             # Valid log levels.
 
@@ -327,310 +385,182 @@ sub initialize( $ ) {
 	( $product, $Product, $toolname, $toolNAME ) = qw( shorewall6 Shorewall6 ip6tables IP6TABLES );
     }
 
-    $verbosity = 0;            # Verbosity setting. -1 = silent, 0 = almost silent, 1 = major progress messages only, 2 = all progress messages (very noisy)
-    $log = undef;              # File reference for log file
-    $log_verbosity = -1;       # Verbosity of log.
-    $timestamp = '';           # If true, we are to timestamp each progress message
-    $script = 0;               # Script (output) file Handle Reference
+    $verbosity      = 0;       # Verbosity setting. -1 = silent, 0 = almost silent, 1 = major progress messages only, 2 = all progress messages (very noisy)
+    $log            = undef;   # File reference for log file
+    $log_verbosity  = -1;      # Verbosity of log.
+    $timestamp      = '';      # If true, we are to timestamp each progress message
+    $script         = 0;       # Script (output) file Handle Reference
     $script_enabled = 0;       # Writing to output file is disabled initially
-    $lastlineblank = 0;        # Avoid extra blank lines in the output
-    $indent1       = '';       # Current indentation tabs
-    $indent2       = '';       # Current indentation spaces
-    $indent        = '';       # Current total indentation
+    $lastlineblank  = 0;       # Avoid extra blank lines in the output
+    $indent1        = '';      # Current indentation tabs
+    $indent2        = '';      # Current indentation spaces
+    $indent         = '';      # Current total indentation
     ( $dir, $file ) = ('',''); # Script's Directory and Filename
-    $tempfile = '';            # Temporary File Name
-    $sillyname = '';           # Temporary ipchain
+    $tempfile       = '';      # Temporary File Name
+    $sillyname      = 
+    $sillyname1     = '';      # Temporary ipchains
 
     #
     # Misc Globals
     #
-    %globals  =   ( SHAREDIR => '/usr/share/shorewall' ,
-		    SHAREDIRPL => '/usr/share/shorewall/' ,
-		    CONFDIR =>  '/etc/shorewall',     # Run-time configuration directory
-		    CONFIGDIR => '',                  # Compile-time configuration directory (location of $product.conf)
-		    LOGPARMS => '',
-		    TC_SCRIPT => '',
-		    EXPORT => 0,
+    %globals  =   ( SHAREDIRPL => '/usr/share/shorewall/' ,
+		    CONFDIR    =>  '/etc/shorewall',     # Run-time configuration directory
+		    CONFIGDIR  => '',                  # Compile-time configuration directory (location of $product.conf)
+		    LOGPARMS   => '',
+		    TC_SCRIPT  => '',
+		    EXPORT     => 0,
 		    STATEMATCH => '-m state --state',
-		    UNTRACKED => 0,
-		    VERSION => "4.4.15.1",
-		    CAPVERSION => 40415 ,
+		    UNTRACKED  => 0,
+		    VERSION    => "4.4.19.1",
+		    CAPVERSION => 40417 ,
 		  );
-
     #
     # From shorewall.conf file
     #
-    if ( $family == F_IPV4 ) {
-	$globals{PRODUCT} = 'shorewall';
+    %config =
+	( STARTUP_ENABLED => undef,
+	  VERBOSITY => undef,
+	  #
+	  # Logging
+	  #
+	  LOGFILE => undef,
+	  LOGFORMAT => undef,
+	  LOGTAGONLY => undef,
+	  LOGLIMIT => undef,
+	  LOGRATE => undef,
+	  LOGBURST => undef,
+	  LOGALLNEW => undef,
+	  BLACKLIST_LOGLEVEL => undef,
+	  RFC1918_LOG_LEVEL => undef,
+	  MACLIST_LOG_LEVEL => undef,
+	  TCP_FLAGS_LOG_LEVEL => undef,
+	  SMURF_LOG_LEVEL => undef,
+	  LOG_MARTIANS => undef,
+	  LOG_VERBOSITY => undef,
+	  STARTUP_LOG => undef,
+	  #
+	  # Location of Files
+	  #
+	  IP => undef,
+	  TC => undef,
+	  IPSET => undef,
+	  PERL => undef,
+	  #
+	  #PATH is inherited
+	  #
+	  PATH => undef,
+	  SHOREWALL_SHELL => undef,
+	  SUBSYSLOCK => undef,
+	  MODULESDIR => undef,
+	  #
+	  #CONFIG_PATH is inherited
+	  #
+	  CONFIG_PATH => undef,
+	  RESTOREFILE => undef,
+	  IPSECFILE => undef,
+	  LOCKFILE => undef,
+	  #
+	  # Default Actions/Macros
+	  #
+	  DROP_DEFAULT => undef,
+	  REJECT_DEFAULT => undef,
+	  ACCEPT_DEFAULT => undef,
+	  QUEUE_DEFAULT => undef,
+	  NFQUEUE_DEFAULT => undef,
+	  #
+	  # RSH/RCP Commands
+	  #
+	  RSH_COMMAND => undef,
+	  RCP_COMMAND => undef,
+	  #
+	  # Firewall Options
+	  #
+	  BRIDGING => undef,
+	  IP_FORWARDING => undef,
+	  ADD_IP_ALIASES => undef,
+	  ADD_SNAT_ALIASES => undef,
+	  RETAIN_ALIASES => undef,
+	  TC_ENABLED => undef,
+	  TC_EXPERT => undef,
+	  TC_PRIOMAP => undef,
+	  CLEAR_TC => undef,
+	  MARK_IN_FORWARD_CHAIN => undef,
+	  CLAMPMSS => undef,
+	  ROUTE_FILTER => undef,
+	  DETECT_DNAT_IPADDRS => undef,
+	  MUTEX_TIMEOUT => undef,
+	  ADMINISABSENTMINDED => undef,
+	  BLACKLISTNEWONLY => undef,
+	  DELAYBLACKLISTLOAD => undef,
+	  MODULE_SUFFIX => undef,
+	  DISABLE_IPV6 => undef,
+	  DYNAMIC_ZONES => undef,
+	  PKTTYPE=> undef,
+	  MACLIST_TABLE => undef,
+	  MACLIST_TTL => undef,
+	  SAVE_IPSETS => undef,
+	  MAPOLDACTIONS => undef,
+	  FASTACCEPT => undef,
+	  IMPLICIT_CONTINUE => undef,
+	  HIGH_ROUTE_MARKS => undef,
+	  USE_ACTIONS=> undef,
+	  OPTIMIZE => undef,
+	  EXPORTPARAMS => undef,
+	  SHOREWALL_COMPILER => undef,
+	  EXPAND_POLICIES => undef,
+	  KEEP_RT_TABLES => undef,
+	  DELETE_THEN_ADD => undef,
+	  MULTICAST => undef,
+	  DONT_LOAD => '',
+	  AUTO_COMMENT => undef ,
+	  MANGLE_ENABLED => undef ,
+	  RFC1918_STRICT => undef ,
+	  NULL_ROUTE_RFC1918 => undef ,
+	  USE_DEFAULT_RT => undef ,
+	  RESTORE_DEFAULT_ROUTE => undef ,
+	  FAST_STOP => undef ,
+	  AUTOMAKE => undef ,
+	  WIDE_TC_MARKS => undef,
+	  TRACK_PROVIDERS => undef,
+	  ZONE2ZONE => undef,
+	  ACCOUNTING => undef,
+	  OPTIMIZE_ACCOUNTING => undef,
+	  DYNAMIC_BLACKLIST => undef,
+	  LOAD_HELPERS_ONLY => undef,
+	  REQUIRE_INTERFACE => undef,
+	  FORWARD_CLEAR_MARK => undef,
+	  COMPLETE => undef,
+	  EXPORTMODULES => undef,
+	  #
+	  # Packet Disposition
+	  #
+	  MACLIST_DISPOSITION => undef,
+	  TCP_FLAGS_DISPOSITION => undef,
+	  BLACKLIST_DISPOSITION => undef,
+	  #
+	  # Mark Geometry
+	  #
+	  TC_BITS => undef,
+	  PROVIDER_BITS => undef,
+	  PROVIDER_OFFSET => undef,
+	  MASK_BITS => undef
+	);
 
-	%config =
-	    ( STARTUP_ENABLED => undef,
-	      VERBOSITY => undef,
-	      #
-	      # Logging
-	      #
-	      LOGFILE => undef,
-	      LOGFORMAT => undef,
-	      LOGTAGONLY => undef,
-	      LOGLIMIT => undef,
-	      LOGRATE => undef,
-	      LOGBURST => undef,
-	      LOGALLNEW => undef,
-	      BLACKLIST_LOGLEVEL => undef,
-	      RFC1918_LOG_LEVEL => undef,
-	      MACLIST_LOG_LEVEL => undef,
-	      TCP_FLAGS_LOG_LEVEL => undef,
-	      SMURF_LOG_LEVEL => undef,
-	      LOG_MARTIANS => undef,
-	      LOG_VERBOSITY => undef,
-	      STARTUP_LOG => undef,
-	      #
-	      # Location of Files
-	      #
-	      IPTABLES => undef,
-	      IP => undef,
-	      TC => undef,
-	      IPSET => undef,
-	      PERL => undef,
-	      #
-	      #PATH is inherited
-	      #
-	      PATH => undef,
-	      SHOREWALL_SHELL => undef,
-	      SUBSYSLOCK => undef,
-	      MODULESDIR => undef,
-	      #
-	      #CONFIG_PATH is inherited
-	      #
-	      CONFIG_PATH => undef,
-	      RESTOREFILE => undef,
-	      IPSECFILE => undef,
-	      LOCKFILE => undef,
-	      #
-	      # Default Actions/Macros
-	      #
-	      DROP_DEFAULT => undef,
-	      REJECT_DEFAULT => undef,
-	      ACCEPT_DEFAULT => undef,
-	      QUEUE_DEFAULT => undef,
-	      NFQUEUE_DEFAULT => undef,
-	      #
-	      # RSH/RCP Commands
-	      #
-	      RSH_COMMAND => undef,
-	      RCP_COMMAND => undef,
-	      #
-	      # Firewall Options
-	      #
-	      BRIDGING => undef,
-	      IP_FORWARDING => undef,
-	      ADD_IP_ALIASES => undef,
-	      ADD_SNAT_ALIASES => undef,
-	      RETAIN_ALIASES => undef,
-	      TC_ENABLED => undef,
-	      TC_EXPERT => undef,
-	      TC_PRIOMAP => undef,
-	      CLEAR_TC => undef,
-	      MARK_IN_FORWARD_CHAIN => undef,
-	      CLAMPMSS => undef,
-	      ROUTE_FILTER => undef,
-	      DETECT_DNAT_IPADDRS => undef,
-	      MUTEX_TIMEOUT => undef,
-	      ADMINISABSENTMINDED => undef,
-	      BLACKLISTNEWONLY => undef,
-	      DELAYBLACKLISTLOAD => undef,
-	      MODULE_SUFFIX => undef,
-	      DISABLE_IPV6 => undef,
-	      DYNAMIC_ZONES => undef,
-	      PKTTYPE=> undef,
-	      MACLIST_TABLE => undef,
-	      MACLIST_TTL => undef,
-	      SAVE_IPSETS => undef,
-	      MAPOLDACTIONS => undef,
-	      FASTACCEPT => undef,
-	      IMPLICIT_CONTINUE => undef,
-	      HIGH_ROUTE_MARKS => undef,
-	      USE_ACTIONS=> undef,
-	      OPTIMIZE => undef,
-	      EXPORTPARAMS => undef,
-	      SHOREWALL_COMPILER => undef,
-	      EXPAND_POLICIES => undef,
-	      KEEP_RT_TABLES => undef,
-	      DELETE_THEN_ADD => undef,
-	      MULTICAST => undef,
-	      DONT_LOAD => '',
-	      AUTO_COMMENT => undef ,
-	      MANGLE_ENABLED => undef ,
-	      RFC1918_STRICT => undef ,
-	      NULL_ROUTE_RFC1918 => undef ,
-	      USE_DEFAULT_RT => undef ,
-	      RESTORE_DEFAULT_ROUTE => undef ,
-	      FAST_STOP => undef ,
-	      AUTOMAKE => undef ,
-	      WIDE_TC_MARKS => undef,
-	      TRACK_PROVIDERS => undef,
-	      ZONE2ZONE => undef,
-	      ACCOUNTING => undef,
-	      OPTIMIZE_ACCOUNTING => undef,
-	      DYNAMIC_BLACKLIST => undef,
-	      LOAD_HELPERS_ONLY => undef,
-	      REQUIRE_INTERFACE => undef,
-	      FORWARD_CLEAR_MARK => undef,
-	      COMPLETE => undef,
-	      #
-	      # Packet Disposition
-	      #
-	      MACLIST_DISPOSITION => undef,
-	      TCP_FLAGS_DISPOSITION => undef,
-	      BLACKLIST_DISPOSITION => undef,
-	      #
-	      # Mark Geometry
-	      #
-	      TC_BITS => undef,
-	      PROVIDER_BITS => undef,
-	      PROVIDER_OFFSET => undef,
-	      MASK_BITS => undef
-	    );
+    %validlevels = ( DEBUG   => 7,
+		     INFO    => 6,
+		     NOTICE  => 5,
+		     WARNING => 4,
+		     WARN    => 4,
+		     ERR     => 3,
+		     ERROR   => 3,
+		     CRIT    => 2,
+		     ALERT   => 1,
+		     EMERG   => 0,
+		     PANIC   => 0,
+		     NONE    => '',
+		     NFLOG   => 'NFLOG',
+		     LOGMARK => 'LOGMARK' );
 
-	%validlevels = ( DEBUG   => 7,
-			 INFO    => 6,
-			 NOTICE  => 5,
-			 WARNING => 4,
-			 WARN    => 4,
-			 ERR     => 3,
-			 ERROR   => 3,
-			 CRIT    => 2,
-			 ALERT   => 1,
-			 EMERG   => 0,
-			 PANIC   => 0,
-			 NONE    => '',
-			 ULOG    => 'ULOG',
-			 NFLOG   => 'NFLOG',
-		         LOGMARK => 'LOGMARK' );
-    } else {
-	$globals{SHAREDIR} = '/usr/share/shorewall6';
-	$globals{CONFDIR}  = '/etc/shorewall6';
-	$globals{PRODUCT}  = 'shorewall6';
-
-	%config =
-	    ( STARTUP_ENABLED => undef,
-	      VERBOSITY => undef,
-	      #
-	      # Logging
-	      #
-	      LOGFILE => undef,
-	      LOGFORMAT => undef,
-	      LOGTAGONLY => undef,
-	      LOGLIMIT => undef,
-	      LOGRATE => undef,
-	      LOGBURST => undef,
-	      LOGALLNEW => undef,
-	      BLACKLIST_LOGLEVEL => undef,
-	      TCP_FLAGS_LOG_LEVEL => undef,
-	      SMURF_LOG_LEVEL => undef,
-	      LOG_VERBOSITY => undef,
-	      STARTUP_LOG => undef,
-	      #
-	      # Location of Files
-	      #
-	      IP6TABLES => undef,
-	      IP => undef,
-	      TC => undef,
-	      IPSET => undef,
-	      PERL => undef,
-	      #
-	      #PATH is inherited
-	      #
-	      PATH => undef,
-	      SHOREWALL_SHELL => undef,
-	      SUBSYSLOCK => undef,
-	      MODULESDIR => undef,
-	      #
-	      #CONFIG_PATH is inherited
-	      #
-	      CONFIG_PATH => undef,
-	      RESTOREFILE => undef,
-	      LOCKFILE => undef,
-	      #
-	      # Default Actions/Macros
-	      #
-	      DROP_DEFAULT => undef,
-	      REJECT_DEFAULT => undef,
-	      ACCEPT_DEFAULT => undef,
-	      QUEUE_DEFAULT => undef,
-	      NFQUEUE_DEFAULT => undef,
-	      #
-	      # RSH/RCP Commands
-	      #
-	      RSH_COMMAND => undef,
-	      RCP_COMMAND => undef,
-	      #
-	      # Firewall Options
-	      #
-	      IP_FORWARDING => undef,
-	      TC_ENABLED => undef,
-	      TC_EXPERT => undef,
-	      TC_PRIOMAP => undef,
-	      CLEAR_TC => undef,
-	      MARK_IN_FORWARD_CHAIN => undef,
-	      CLAMPMSS => undef,
-	      MUTEX_TIMEOUT => undef,
-	      ADMINISABSENTMINDED => undef,
-	      BLACKLISTNEWONLY => undef,
-	      MODULE_SUFFIX => undef,
-	      MAPOLDACTIONS => '',
-	      FASTACCEPT => undef,
-	      IMPLICIT_CONTINUE => undef,
-	      HIGH_ROUTE_MARKS => undef,
-	      OPTIMIZE => undef,
-	      EXPORTPARAMS => undef,
-	      EXPAND_POLICIES => undef,
-	      KEEP_RT_TABLES => undef,
-	      DELETE_THEN_ADD => undef,
-	      MULTICAST => undef,
-	      DONT_LOAD => '',
-	      AUTO_COMMENT => undef,
-	      MANGLE_ENABLED => undef ,
-	      AUTOMAKE => undef ,
-	      WIDE_TC_MARKS => undef,
-	      TRACK_PROVIDERS => undef,
-	      ZONE2ZONE => undef,
-	      ACCOUNTING => undef,
-	      OPTIMIZE_ACCOUNTING => undef,
-	      DYNAMIC_BLACKLIST => undef,
-	      LOAD_HELPERS_ONLY => undef,
-	      REQUIRE_INTERFACE => undef,
-	      FORWARD_CLEAR_MARK => undef,
-	      COMPLETE => undef,
-	      #
-	      # Packet Disposition
-	      #
-	      TCP_FLAGS_DISPOSITION => undef,
-	      BLACKLIST_DISPOSITION => undef,
-	      #
-	      # Mark Geometry
-	      #
-	      TC_BITS => undef,
-	      PROVIDER_BITS => undef,
-	      PROVIDER_OFFSET => undef,
-	      MASK_BITS => undef
-	    );
-
-	%validlevels = ( DEBUG   => 7,
-			 INFO    => 6,
-			 NOTICE  => 5,
-			 WARNING => 4,
-			 WARN    => 4,
-			 ERR     => 3,
-			 ERROR   => 3,
-			 CRIT    => 2,
-			 ALERT   => 1,
-			 EMERG   => 0,
-			 PANIC   => 0,
-			 NONE    => '',
-			 NFLOG   => 'NFLOG',
-		         LOGMARK => 'LOGMARK' );
-    }
     #
     # From parsing the capabilities file or capabilities detection
     #
@@ -686,6 +616,7 @@ sub initialize( $ ) {
 	       FWMARK_RT_MASK => undef,
 	       MARK_ANYWHERE => undef,
 	       HEADER_MATCH => undef,
+               ACCOUNT_TARGET => undef,
 	       CAPVERSION => undef,
 	       KERNELVERSION => undef,
 	       );
@@ -711,17 +642,30 @@ sub initialize( $ ) {
     $shorewall_dir = '';      #Shorewall Directory
 
     $debug = 0;
+    
+    %params = ( root        => '',
+		system      => '',
+		command     => '',
+		files       => '',
+		destination => '' );
 
-    %params = ();
-}
+    %compiler_params = ();
 
-INIT {
-    #
-    # These variables appear within single quotes in shorewall.conf -- add them to ENV
-    # so that read_a_line doesn't have to be smart enough to parse that usage.
-    #
-    for ( qw/root system command files destination/ ) {
-	$ENV{$_} = '' unless exists $ENV{$_};
+    $compiler_params{$_} = 1 for keys %params;
+
+    %actparms = ();
+
+    if ( $family == F_IPV4 ) {
+	$globals{SHAREDIR} = '/usr/share/shorewall';
+	$globals{CONFDIR}  = '/etc/shorewall';
+	$globals{PRODUCT}  = 'shorewall';
+	$config{IPTABLES}  = undef;
+	$validlevels{ULOG} = 'ULOG',
+    } else {
+	$globals{SHAREDIR} = '/usr/share/shorewall6';
+	$globals{CONFDIR}  = '/etc/shorewall6';
+	$globals{PRODUCT}  = 'shorewall6';
+	$config{IP6TABLES} = undef;
     }
 }
 
@@ -812,7 +756,7 @@ sub fatal_error	{
     die "   ERROR: @_$currentlineinfo\n";
 }
 
-sub fatal_error1	{
+sub fatal_error1 {
     $| = 1;
 
     if ( $log ) {
@@ -1152,190 +1096,6 @@ sub pop_indent() {
     }
 }
 
-#
-# Functions for copying files into the script
-#
-sub copy( $ ) {
-    assert( $script_enabled );
-
-    if ( $script ) {
-	my $file = $_[0];
-
-	open IF , $file or fatal_error "Unable to open $file: $!";
-
-	while ( <IF> ) {
-	    chomp;
-	    if ( /^\s*$/ ) {
-		print $script "\n" unless $lastlineblank;
-		$lastlineblank = 1;
-	    } else {
-		if  ( $indent ) {
-		    s/^(\s*)/$indent1$1$indent2/;
-		    s/        /\t/ if $indent2;
-		}
-
-		print $script $_;
-		print $script "\n";
-		$lastlineblank = 0;
-	    }
-	}
-
-	close IF;
-    }
-}
-
-#
-# This one handles line continuation and 'here documents'
-
-sub copy1( $ ) {
-    assert( $script_enabled );
-
-    my $result = 0;
-
-    if ( $script || $debug ) {
-	my $file = $_[0];
-
-	open IF , $file or fatal_error "Unable to open $file: $!";
-
-	my ( $do_indent, $here_documents ) = ( 1, '');
-
-	while ( <IF> ) {
-	    chomp;
-
-	    if ( /^${here_documents}\s*$/ ) {
-		if ( $script ) {
-		    print $script $here_documents if $here_documents;
-		    print $script "\n";
-		}
-
-		if ( $debug ) {
-		    print "GS-----> $here_documents" if $here_documents;
-		    print "GS----->\n";
-		}
-
-		$do_indent = 1;
-		$here_documents = '';
-		next;
-	    }
-
-	    if ( $do_indent && /.*<<\s*([^ ]+)s*(.*)/ ) {
-		$here_documents = $1;
-		s/^(\s*)/$indent1$1$indent2/;
-		s/        /\t/ if $indent2;
-		$do_indent = 0;
-
-		if ( $script ) {
-		    print $script $_;
-		    print $script "\n";
-		}
-
-		if ( $debug ) {
-		    s/\n/\nGS-----> /g;
-		    print "GS-----> $_\n";
-		}
-
-		$result = 1;
-		next;
-	    }
-
-	    if ( $indent && $do_indent ) {
-		s/^(\s*)/$indent1$1$indent2/;
-		s/        /\t/ if $indent2;
-	    }
-
-	    if ( $script ) {
-		print $script $_;
-		print $script "\n";
-	    }
-
-	    $do_indent = ! ( $here_documents || /\\$/ );
-
-	    $result = 1 unless $result || /^\s*$/ || /^\s*#/;
-
-	    if ( $debug ) {
-		s/\n/\nGS-----> /g;
-		print "GS-----> $_\n";
-	    }
-	}
-
-	close IF;
-    }
-
-    $lastlineblank = 0;
-
-    $result;
-}
-
-#
-# This one drops header comments and replaces them with a three-line banner
-#
-sub copy2( $$ ) {
-    my ( $file, $trace ) = @_;
-
-    assert( $script_enabled );
-    my $empty = 1;
-
-    if ( $script || $trace ) {
-	my $file = $_[0];
-
-	open IF , $file or fatal_error "Unable to open $file: $!";
-
-	while ( <IF> ) {
-	    $empty = 0, last unless /^#/;
-	}
-
-	unless ( $empty ) {
-	    emit <<EOF;
-################################################################################
-#   Functions imported from $file
-################################################################################
-EOF
-	    chomp;
-	    emit( $_ ) unless /^\s*$/;
-
-	    while ( <IF> ) {
-		chomp;
-		if ( /^\s*$/ ) {
-		    unless ( $lastlineblank ) {
-			print $script "\n" if $script;
-			print "GS----->\n" if $trace;
-		    }
-
-		    $lastlineblank = 1;
-		} else {
-		    if  ( $indent ) {
-			s/^(\s*)/$indent1$1$indent2/;
-			s/        /\t/ if $indent2;
-		    }
-
-		    if ( $script ) {
-			print $script $_;
-			print $script "\n";
-		    }
-
-		    if ( $trace ) {
-			s/\n/GS-----> \n/g;
-			print "GS-----> $_\n";
-		    }
-
-		    $lastlineblank = 0;
-		}
-	    }
-
-	    close IF;
-
-	    unless ( $lastlineblank ) {
-		print $script "\n" if $script;
-		print "GS----->\n" if $trace;
-	    }
-
-	    emit( '################################################################################',
-		  "#   End of imports from $file",
-		  '################################################################################' );
-	}
-    }
-}
-
 #
 # Create the temporary script file -- the passed file name is the name of the final file.
 # We create a temporary file in the same directory so that we can use rename to finalize it.
@@ -1611,7 +1371,6 @@ sub pop_include() {
 # This function is normally called below in read_a_line() when EOF is reached. Clients of the
 # module may also call the function to close the file before EOF
 #
-
 sub close_file() {
     if ( $currentfile ) {
 	my $result = close $currentfile;
@@ -1625,9 +1384,219 @@ sub close_file() {
     }
 }
 
+#
+# Functions for copying files into the script
+#
+sub copy( $ ) {
+    assert( $script_enabled );
+
+    if ( $script ) {
+	my $file = $_[0];
+
+	open IF , $file or fatal_error "Unable to open $file: $!";
+
+	while ( <IF> ) {
+	    chomp;
+	    if ( /^\s*$/ ) {
+		print $script "\n" unless $lastlineblank;
+		$lastlineblank = 1;
+	    } else {
+		if  ( $indent ) {
+		    s/^(\s*)/$indent1$1$indent2/;
+		    s/        /\t/ if $indent2;
+		}
+
+		print $script $_;
+		print $script "\n";
+		$lastlineblank = 0;
+	    }
+	}
+
+	close IF;
+    }
+}
+
+#
+# This variant of copy handles line continuation, 'here documents' and INCLUDE
+#
+sub copy1( $ ) {
+    assert( $script_enabled );
+
+    my $result = 0;
+
+    if ( $script || $debug ) {
+	my ( $do_indent, $here_documents ) = ( 1, '');
+
+	open_file( $_[0] );
+	
+	while ( $currentfile ) {
+	    while ( <$currentfile> ) {
+		$currentlinenumber++;
+
+		chomp;
+
+		if ( /^${here_documents}\s*$/ ) {
+		    if ( $script ) {
+			print $script $here_documents if $here_documents;
+			print $script "\n";
+		    }
+
+		    if ( $debug ) {
+			print "GS-----> $here_documents" if $here_documents;
+			print "GS----->\n";
+		    }
+
+		    $do_indent = 1;
+		    $here_documents = '';
+		    next;
+		}
+
+		if ( $do_indent && /.*<<\s*([^ ]+)s*(.*)/ ) {
+		    $here_documents = $1;
+		    s/^(\s*)/$indent1$1$indent2/;
+		    s/        /\t/ if $indent2;
+		    $do_indent = 0;
+
+		    if ( $script ) {
+			print $script $_;
+			print $script "\n";
+		    }
+
+		    if ( $debug ) {
+			s/\n/\nGS-----> /g;
+			print "GS-----> $_\n";
+		    }
+
+		    $result = 1;
+		    next;
+		}
+
+		if ( $do_indent ) {
+		    if ( /^\s*INCLUDE\b/ ) {
+			my @line = split / /;
+
+			fatal_error "Invalid INCLUDE command"    if @line != 2;
+			fatal_error "INCLUDEs nested too deeply" if @includestack >= 4;
+
+			my $filename = find_file $line[1];
+
+			fatal_error "INCLUDE file $filename not found" unless -f $filename;
+			fatal_error "Directory ($filename) not allowed in INCLUDE" if -d _;
+
+			if ( -s _ ) {
+			    push @includestack, [ $currentfile, $currentfilename, $currentlinenumber ];
+			    $currentfile = undef;
+			    do_open_file $filename;
+			} else {
+			    $currentlinenumber = 0;
+			}
+
+			next;
+		    }
+			
+		    if ( $indent ) {
+			s/^(\s*)/$indent1$1$indent2/;
+			s/        /\t/ if $indent2;
+		    }
+		}
+
+		if ( $script ) {
+		    print $script $_;
+		    print $script "\n";
+		}
+
+		$do_indent = ! ( $here_documents || /\\$/ );
+
+		$result = 1 unless $result || /^\s*$/ || /^\s*#/;
+
+		if ( $debug ) {
+		    s/\n/\nGS-----> /g;
+		    print "GS-----> $_\n";
+		}
+	    }
+
+	    close_file;
+	}
+    }
+
+    $lastlineblank = 0;
+
+    $result;
+}
+
+#
+# This one drops header comments and replaces them with a three-line banner
+#
+sub copy2( $$ ) {
+    my ( $file, $trace ) = @_;
+
+    assert( $script_enabled );
+    my $empty = 1;
+
+    if ( $script || $trace ) {
+	my $file = $_[0];
+
+	open IF , $file or fatal_error "Unable to open $file: $!";
+
+	while ( <IF> ) {
+	    $empty = 0, last unless /^#/;
+	}
+
+	unless ( $empty ) {
+	    emit <<EOF;
+################################################################################
+#   Functions imported from $file
+################################################################################
+EOF
+	    chomp;
+	    emit( $_ ) unless /^\s*$/;
+
+	    while ( <IF> ) {
+		chomp;
+		if ( /^\s*$/ ) {
+		    unless ( $lastlineblank ) {
+			print $script "\n" if $script;
+			print "GS----->\n" if $trace;
+		    }
+
+		    $lastlineblank = 1;
+		} else {
+		    if  ( $indent ) {
+			s/^(\s*)/$indent1$1$indent2/;
+			s/        /\t/ if $indent2;
+		    }
+
+		    if ( $script ) {
+			print $script $_;
+			print $script "\n";
+		    }
+
+		    if ( $trace ) {
+			s/\n/GS-----> \n/g;
+			print "GS-----> $_\n";
+		    }
+
+		    $lastlineblank = 0;
+		}
+	    }
+
+	    close IF;
+
+	    unless ( $lastlineblank ) {
+		print $script "\n" if $script;
+		print "GS----->\n" if $trace;
+	    }
+
+	    emit( '################################################################################',
+		  "#   End of imports from $file",
+		  '################################################################################' );
+	}
+    }
+}
+
 #
 # The following two functions allow module clients to nest opens. This happens frequently
-# in the Actions module.
+# in the Rules module.
 #
 sub push_open( $ ) {
 
@@ -1787,6 +1756,29 @@ sub embedded_perl( $ ) {
     }
 }
 
+#
+# Push/pop action params
+#
+sub push_params( $ ) {
+    my @params = split /,/, $_[0];
+    my $oldparams = \%actparms;
+
+    %actparms = ();
+
+    for ( my $i = 1; $i <= @params; $i++ ) {
+	my $val = $params[$i - 1];
+
+	$actparms{$i} = $val eq '-' ? '' : $val eq '--' ? '-' : $val;
+    }
+
+    $oldparams;
+}
+
+sub pop_params( $ ) {
+    my $oldparms = shift;
+    %actparms = %$oldparms;
+}
+
 #
 # Read a line from the current include stack.
 #
@@ -1860,26 +1852,25 @@ sub read_a_line(;$) {
 
 	    my $count = 0;
 	    #
-	    # Expand Shell Variables using %params and %ENV
+	    # Expand Shell Variables using %params and %actparms
 	    #
-	    #                            $1      $2      $3           -     $4
-	    while ( $currentline =~ m( ^(.*?) \$({)? ([a-zA-Z]\w*) (?(2)}) (.*)$ )x ) {
+	    #                            $1      $2   $3      -     $4
+	    while ( $currentline =~ m( ^(.*?) \$({)? (\w+) (?(2)}) (.*)$ )x ) {
 
-		unless ( exists $params{$3} ) {
-		    #
-		    # Given the way that getparams works, this should never help but better safe than sorry
-		    #
-		    $params{$3} = $ENV{$3} if exists $ENV{$3};
+		my ( $first, $var, $rest ) = ( $1, $3, $4);
+
+		my $val;
+
+		if ( $var =~ /^\d+$/ ) {
+		    fatal_error "Undefined parameter (\$$var)" unless exists $actparms{$var};
+		    $val = $actparms{$var};
+		} else {
+		    fatal_error "Undefined shell variable (\$$var)" unless exists $params{$var};
+		    $val = $params{$var};
 		}
 
-		my $val = $params{$3};
-
-		unless ( defined $val ) {
-		    fatal_error "Undefined shell variable (\$$3)" unless exists $params{$3} || exists $ENV{$3};
-		    $val = '';
-		}
-
-		$currentline = join( '', $1 , $val , $4 );
+		$val = '' unless defined $val;
+		$currentline = join( '', $first , $val , $rest );
 		fatal_error "Variable Expansion Loop" if ++$count > 100;
 	    }
 
@@ -2076,9 +2067,10 @@ sub default_log_level( $$ ) {
 #
 sub check_trivalue( $$ ) {
     my ( $var, $default) = @_;
-    my $val = lc( $config{$var} || '' );
+    my $val = $config{$var};
 
     if ( defined $val ) {
+	$val = lc $val;
 	if ( $val eq 'yes' || $val eq 'on' ) {
 	    $config{$var} = 'on';
 	} elsif ( $val eq 'no' || $val eq 'off' ) {
@@ -2091,7 +2083,7 @@ sub check_trivalue( $$ ) {
 	    fatal_error "Invalid value ($val) for $var";
 	}
     } else {
-	$config{var} = $default
+	$config{$var} = $default
     }
 }
 
@@ -2144,12 +2136,16 @@ sub load_kernel_modules( ) {
 	my $uname = `uname -r`;
 	fatal_error "The command 'uname -r' failed" unless $? == 0;
 	chomp $uname;
-	$modulesdir = "/lib/modules/$uname/kernel/net/ipv4/netfilter:/lib/modules/$uname/kernel/net/netfilter:/lib/modules/$uname/extra:/lib/modules/$uname/extra/ipset";
+	$modulesdir = "/lib/modules/$uname/kernel/net/ipv4/netfilter:/lib/modules/$uname/kernel/net/ipv6/netfilter:/lib/modules/$uname/kernel/net/netfilter:/lib/modules/$uname/extra:/lib/modules/$uname/extra/ipset";
     }
 
-    my @moduledirectories = split /:/, $modulesdir;
+    my @moduledirectories; 
 
-    if ( $moduleloader && open_file( $config{LOAD_HELPERS_ONLY} ? 'helpers' : 'modules' ) ) {
+    for ( split /:/, $modulesdir ) {
+	push @moduledirectories, $_ if -d $_;
+    }
+
+    if ( $moduleloader &&  @moduledirectories && open_file( $config{LOAD_HELPERS_ONLY} ? 'helpers' : 'modules' ) ) {
 	my %loadedmodules;
 
 	$loadedmodules{$_}++ for split_list( $config{DONT_LOAD}, 'module' );
@@ -2182,7 +2178,7 @@ sub load_kernel_modules( ) {
 			    } else {
 				system( "modprobe $module $arguments" );
 			    }
-
+				    
 			    $loadedmodules{ $module } = 1;
 			}
 		    }
@@ -2509,8 +2505,17 @@ sub Header_Match() {
     qt1( "$iptables -A $sillyname -m ipv6header --header 255 -j ACCEPT" );
 }
 
+sub Account_Target() {
+    if ( $family == F_IPV4 ) {
+	qt1( "$iptables -A $sillyname -j ACCOUNT --addr 192.168.1.0/29 --tname $sillyname" );
+    } else {
+	qt1( "$iptables -A $sillyname -j ACCOUNT --addr 1::/122 --tname $sillyname" );
+    }
+}
+
 our %detect_capability =
-    ( ADDRTYPE => \&Addrtype,
+    ( ACCOUNT_TARGET =>\&Account_Target,
+      ADDRTYPE => \&Addrtype,
       CLASSIFY_TARGET => \&Classify_Target,
       COMMENTS => \&Comments,
       CONNLIMIT_MATCH => \&Connlimit_Match,
@@ -2683,6 +2688,7 @@ sub determine_capabilities() {
 	$capabilities{FLOW_FILTER}     = detect_capability( 'FLOW_FILTER' );
 	$capabilities{FWMARK_RT_MASK}  = detect_capability( 'FWMARK_RT_MASK' );
 	$capabilities{MARK_ANYWHERE}   = detect_capability( 'MARK_ANYWHERE' );
+	$capabilities{ACCOUNT_TARGET}  = detect_capability( 'ACCOUNT_TARGET' );
 
 
 	qt1( "$iptables -F $sillyname" );
@@ -2717,7 +2723,7 @@ sub ensure_config_path() {
 
 	open_file $f;
 
-	$ENV{CONFDIR} = $globals{CONFDIR};
+	add_param( CONFDIR => $globals{CONFDIR} );
 
 	while ( read_a_line ) {
 	    if ( $currentline =~ /^\s*([a-zA-Z]\w*)=(.*?)\s*$/ ) {
@@ -2900,17 +2906,145 @@ sub get_params() {
     if ( -f $fn ) {
 	progress_message2 "Processing $fn ...";
 
-	my $command = "$globals{SHAREDIRPL}/getparams $fn " . join( ':', @config_path );
-
+	my $command = "$FindBin::Bin/getparams $fn " . join( ':', @config_path );
+	#
+	# getparams silently sources the params file under 'set -a', then executes 'export -p'
+	#
 	my @params = `$command`;
 
 	fatal_error "Processing of $fn failed" if $?;
 
-	for ( @params ) {
-	    if ( /^(.*?)=(.*)$/ ) {
-		$params{$1} = $2 unless $1 eq '_';
-	    }	
-	} 
+	if ( $debug ) {
+	    print "Params:\n";
+	    print $_ for @params;
+	}
+	
+	my ( $variable , $bug );
+
+	if ( $params[0] =~ /^declare/ ) {
+	    #
+	    # getparams was interpreted by bash
+	    #
+	    # - Variable names are preceded by 'declare -x '
+	    # - Param values are delimited by double quotes
+	    # - Embedded double quotes are escaped with '\\'
+	    # - Valueless variables are supported (e.g., 'declare -x foo')
+	    #
+	    for ( @params ) {
+		if ( /^declare -x (.*?)="(.*[^\\])"$/ ) {
+		    $params{$1} = $2 unless $1 eq '_';
+		} elsif ( /^declare -x (.*?)="(.*)$/ ) {
+		    $params{$variable=$1} = $2 eq '"' ? '' : "${2}\n";
+		} elsif ( /^declare -x (.*)\s+$/ || /^declare -x (.*)=""$/ ) {
+		    $params{$1} = '';
+		} else {
+		    if ($variable) {
+			s/"$//;
+			$params{$variable} .= $_;
+		    } else {
+			chomp;
+			warning_message "Param line ($_) ignored" unless $bug++;
+		    }
+		}	
+	    }
+	} elsif ( $params[0] =~ /^export .*?="/ || $params[0] =~ /^export [^\s=]+\s*$/ ) {
+	    #
+	    # getparams interpreted by older (e.g., RHEL 5) Bash
+	    #
+	    # - Variable names preceded by 'export '
+	    # - Variable values are delimited by double quotes
+	    # - Embedded single quotes are escaped with '\'
+	    # - Valueless variables ( e.g., 'export foo') are supported
+	    #
+	    for ( @params ) {
+		if ( /^export (.*?)="(.*[^\\])"$/ ) {
+		    $params{$1} = $2 unless $1 eq '_';
+		} elsif ( /^export (.*?)="(.*)$/ ) {
+		    $params{$variable=$1} = $2 eq '"' ? '' : "${2}\n";
+		} elsif ( /^export ([^\s=]+)\s*$/ || /^export (.*)=""$/ ) {
+		    $params{$1} = '';
+		} else {
+		    if ($variable) {
+			s/"$//;
+			$params{$variable} .= $_;
+		    } else {
+			chomp;
+			warning_message "Param line ($_) ignored" unless $bug++;
+		    }
+		}	
+	    }
+	} else {
+	    #
+	    # getparams was interpreted by dash/ash/busybox
+	    #
+	    # - Variable name preceded by 'export '
+	    # - Param values are delimited by single quotes.
+	    # - Embedded single quotes are transformed to the five characters '"'"'
+	    #
+	    for ( @params ) {
+		if ( /^export (.*?)='(.*'"'"')$/ ) {
+		    $params{$variable=$1}="${2}\n";		    
+		} elsif ( /^export (.*?)='(.*)'$/ ) {
+		    $params{$1} = $2 unless $1 eq '_';
+		} elsif ( /^export (.*?)='(.*)$/ ) {
+		    $params{$variable=$1}="${2}\n";
+		} else {
+		    if ($variable) {
+			s/'$//;
+			$params{$variable} .= $_;
+		    } else {
+			chomp;
+			warning_message "Param line ($_) ignored" unless $bug++;
+		    }				
+		}
+	    }
+	}
+
+	if ( $debug ) {
+	    print "PARAMS:\n";
+	    my $value;
+	    while ( ($variable, $value ) = each %params ) {
+		print "   $variable='$value'\n" unless $compiler_params{$variable};
+	    }
+	}
+    }
+}
+
+#
+# Add an entry to %params and to %compiler_params
+#
+sub add_param( $$ ) {
+    my ( $param, $value ) = @_;
+
+    $params{$param} = $value;
+    $compiler_params{$param} = 1;
+}
+
+#
+# emit param=value for each param set in the params file
+#
+sub export_params() {
+    my $count = 0;
+
+    while ( my ( $param, $value ) = each %params ) {
+	#
+	# Don't export params added by the compiler
+	#
+	next if exists $compiler_params{$param};
+	#
+	# Don't export pairs from %ENV
+	#
+	if ( exists $ENV{$param} && defined $ENV{$param} ) {
+	    next if $value eq $ENV{$param};
+	}
+
+	emit "#\n# From the params file\n#" unless $count++;
+
+	if ( $value =~ /[\s()[]/ ) {
+	    emit "$param='$value'";
+	} else {
+	    emit "$param=$value";
+	}
     }
 }
 
@@ -2945,11 +3079,12 @@ sub get_configuration( $ ) {
     #
     # get_capabilities requires that the true settings of these options be established
     #
-    default 'MODULE_PREFIX', 'o gz ko o.gz ko.gz';
+    default 'MODULE_PREFIX', 'ko ko.gz o o.gz gz';
     default_yes_no 'LOAD_HELPERS_ONLY'          , '';
 
     get_capabilities( $export );
 
+
     $globals{STATEMATCH} = '-m conntrack --ctstate' if have_capability 'CONNTRACK_MATCH';
 
     if ( my $rate = $config{LOGLIMIT} ) {
@@ -3137,6 +3272,7 @@ sub get_configuration( $ ) {
     default_yes_no 'REQUIRE_INTERFACE'          , '';
     default_yes_no 'FORWARD_CLEAR_MARK'         , have_capability 'MARK' ? 'Yes' : '';
     default_yes_no 'COMPLETE'                   , '';
+    default_yes_no 'EXPORTMODULES'              , '';
 
     require_capability 'MARK' , 'FOREWARD_CLEAR_MARK=Yes', 's', if $config{FORWARD_CLEAR_MARK};
 
@@ -3174,6 +3310,10 @@ sub get_configuration( $ ) {
 
     default 'BLACKLIST_DISPOSITION'    , 'DROP';
 
+    unless ( $config{BLACKLIST_DISPOSITION} eq 'DROP' || $config{BLACKLIST_DISPOSITION} eq 'REJECT' ) {
+	fatal_error q(BLACKLIST_DISPOSITION must be 'DROP' or 'REJECT');
+    }
+
     default_log_level 'BLACKLIST_LOGLEVEL',  '';
     default_log_level 'MACLIST_LOG_LEVEL',   '';
     default_log_level 'TCP_FLAGS_LOG_LEVEL', '';
@@ -3244,7 +3384,7 @@ sub get_configuration( $ ) {
 	my @priomap = split ' ',$val;
 	fatal_error "Invalid TC_PRIOMAP ($val)" unless @priomap == 16;
 	for ( @priomap ) {
-	    fatal_error "Invalid TC_PRIOMAP entry ($_)" unless /[1-3]/;
+	    fatal_error "Invalid TC_PRIOMAP entry ($_)" unless /^[1-3]$/;
 	    $_--;
 	}
 
@@ -3366,7 +3506,7 @@ sub run_user_exit( $ ) {
     my $file = find_file $chainref->{name};
 
     if ( -f $file ) {
-	progress_message2 "Processing $file...";
+	progress_message2 "Running $file...";
 
 	my $command = qq(package Shorewall::User;\nno strict;\n# line 1 "$file"\n) . `cat $file`;
 
@@ -3387,7 +3527,7 @@ sub run_user_exit1( $ ) {
     my $file = find_file $_[0];
 
     if ( -f $file ) {
-	progress_message2 "Processing $file...";
+	progress_message2 "Running $file...";
 	#
 	# File may be empty -- in which case eval would fail
 	#
@@ -3418,7 +3558,7 @@ sub run_user_exit2( $$ ) {
     my ($file, $chainref) = ( find_file $_[0], $_[1] );
 
     if ( -f $file ) {
-	progress_message2 "Processing $file...";
+	progress_message2 "Running $file...";
 	#
 	# File may be empty -- in which case eval would fail
 	#
