@@ -40,7 +40,7 @@ use strict;
 our @ISA = qw(Exporter);
 our @EXPORT = qw( setup_tc );
 our @EXPORT_OK = qw( process_tc_rule initialize );
-our $VERSION = '4.4_15';
+our $VERSION = '4.4_19';
 
 our %tcs = ( T => { chain  => 'tcpost',
 		    connmark => 0,
@@ -476,6 +476,8 @@ sub process_simple_device() {
 
     my $number = in_hexp( $tcdevices{$device} = ++$devnum );
 
+    fatal_error "Unknown interface( $device )" unless known_interface $device;
+
     my $physical = physical_name $device;
     my $dev      = chain_base( $physical );
 
@@ -515,7 +517,8 @@ sub process_simple_device() {
 	 );
 
     emit ( "run_tc qdisc add dev $physical handle ffff: ingress",
-	   "run_tc filter add dev $physical parent ffff: protocol all prio 10 u32 match ip src 0.0.0.0/0 police rate ${in_bandwidth}kbit burst $in_burst drop flowid :1\n"
+	   "run_tc filter add dev $physical parent ffff: protocol all prio 10 u32 match ip src "  . ALLIPv4 . " police rate ${in_bandwidth}kbit burst $in_burst drop flowid :1\n",
+	   "run_tc filter add dev $physical parent ffff: protocol all prio 10 u32 match ip6 src " . ALLIPv6 . " police rate ${in_bandwidth}kbit burst $in_burst drop flowid :1\n"
 	 ) if $in_bandwidth;
 
     if ( $out_part ne '-' ) {
@@ -541,6 +544,8 @@ sub process_simple_device() {
 	    $command .= ' latency 200ms';
 	}
 
+	$command .= ' mpu 64'; #Assume Ethernet
+
 	if ( defined $peak && $peak ne '' ) {
 	    fatal_error "Invalid peak ($peak)" unless $peak =~ /^\d+(?:\.\d+)?(k|kb|m|mb|mbit|kbit|b)?$/;
 	    $command .= " peakrate $peak";
@@ -562,10 +567,13 @@ sub process_simple_device() {
 
     for ( my $i = 1; $i <= 3; $i++ ) {
 	emit "run_tc qdisc add dev $physical parent $number:$i handle ${number}${i}: sfq quantum 1875 limit 127 perturb 10";
-	emit "run_tc filter add dev $physical protocol all parent $number: handle $i fw classid $number:$i";
+	emit "run_tc filter add dev $physical protocol all prio 2 parent $number: handle $i fw classid $number:$i";
 	emit "run_tc filter add dev $physical protocol all prio 1 parent ${number}$i: handle ${number}${i} flow hash keys $type divisor 1024" if $type ne '-' && have_capability 'FLOW_FILTER';
 	emit '';
     }
+    
+    emit "run_tc filter add dev $physical parent $number:0 protocol all prio 1 u32 match ip protocol 6 0xff match u8 0x05 0x0f at 0 match u16 0x0000 0xffc0 at 2 match u8 0x10 0xff at 33 flowid $number:1\n";
+    emit "run_tc filter add dev $physical parent $number:0 protocol all prio 1 u32 match ip6 protocol 6 0xff match u8 0x05 0x0f at 0 match u16 0x0000 0xffc0 at 2 match u8 0x10 0xff at 33 flowid $number:1\n";
 
     save_progress_message_short qq("   TC Device $physical defined.");
 
@@ -644,11 +652,26 @@ sub validate_tc_device( ) {
 	    fatal_error "Invalid device name ($rdevice)" if $rdevice =~ /[:+]/;
 	    my $rdevref = $tcdevices{$rdevice};
 	    fatal_error "REDIRECTED device ($rdevice) has not been defined in this file" unless $rdevref;
-	    fatal_error "IN-BANDWIDTH must be zero for REDIRECTED devices" if $rdevref->{in_bandwidth} ne '0kbit';
+	    fatal_error "IN-BANDWIDTH must be zero for REDIRECTED devices" if $rdevref->{in_bandwidth} != 0;
 	}
     }
 
-    $tcdevices{$device} = { in_bandwidth  => rate_to_kbit( $inband ) . 'kbit',
+    my $in_burst = '10kb';
+
+    if ( $inband =~ /:/ ) {
+	my ( $in_band, $burst ) = split /:/, $inband, 2;
+
+	if ( defined $burst && $burst ne '' ) {
+	    fatal_error "Invalid IN-BANDWIDTH" if $burst =~ /:/;
+	    fatal_error "Invalid burst ($burst)" unless $burst =~ /^\d+(k|kb|m|mb|mbit|kbit|b)?$/;
+	    $in_burst = $burst;
+	}
+
+	$inband = $in_band;
+    }
+
+    $tcdevices{$device} = { in_bandwidth  => rate_to_kbit( $inband ),
+			    in_burst      => $in_burst,
 			    out_bandwidth => rate_to_kbit( $outband ) . 'kbit',
 			    number        => $devnumber,
 			    classify      => $classify,
@@ -755,7 +778,7 @@ sub validate_tc_class( ) {
 	    }
 
 	    fatal_error "Invalid interface/class number ($devclass)" unless defined $classnumber && $classnumber;
-	    fatal_error "Duplicate interface/class number ($devclass)" if defined $devnums[ $classnumber ];
+	    fatal_error "Duplicate interface:class number ($number:$classnumber}" if $tcclasses{$device}{$classnumber};
 	} else {
 	    fatal_error "Missing interface NUMBER";
 	}
@@ -802,7 +825,8 @@ sub validate_tc_class( ) {
 	#
 	$parentref = $tcref->{$parentclass};
 	fatal_error "Unknown Parent class ($parentclass)" unless $parentref && $parentref->{occurs} == 1;
-	fatal_error "The parent class ($parentclass) specifies UMAX and/or DMAX; it cannot serve as a parent" if $parentref->{dmax};
+	fatal_error "The class ($parentclass) specifies UMAX and/or DMAX; it cannot serve as a parent" if $parentref->{dmax};
+	fatal_error "The class ($parentclass) specifies flow; it cannot serve as a parent"             if $parentref->{flow}; 
 	$parentref->{leaf} = 0;
 	$ratemax  = $parentref->{rate};
 	$ratename = q(the parent class's RATE);
@@ -946,6 +970,8 @@ sub process_tc_filter() {
 
     my ($device, $class, $rest ) = split /:/, $devclass, 3;
 
+    our $lastdevice;
+
     fatal_error "Invalid INTERFACE:CLASS ($devclass)" if defined $rest || ! ($device && $class );
 
     my ( $ip, $ip32, $prio , $lo ) = $family == F_IPV4 ? ('ip', 'ip', 10, 2 ) : ('ipv6', 'ip6', 11 , 4 );
@@ -967,6 +993,17 @@ sub process_tc_filter() {
     fatal_error "Unknown CLASS ($devclass)"                  unless $tcref && $tcref->{occurs};
     fatal_error "Filters may not specify an occurring CLASS" if $tcref->{occurs} > 1;
 
+    if ( $devref->{physical} ne $lastdevice ) {
+	if ( $lastdevice ) {
+	    pop_indent;
+	    emit "fi\n";
+	}
+
+	$lastdevice = $devref->{physical};
+	emit "if interface_is_up $lastdevice; then";
+	push_indent;
+    }
+
     my $rule = "filter add dev $devref->{physical} protocol $ip parent $devnum:0 prio $prio u32";
 
     if ( $source ne '-' ) {
@@ -976,7 +1013,7 @@ sub process_tc_filter() {
 
     if ( $dest ne '-' ) {
 	my ( $net , $mask ) = decompose_net( $dest );
-	$rule .= "\\\n   match $ip dst $net/$mask";
+	$rule .= "\\\n   match $ip32 dst $net/$mask";
     }
 
     if ( $tos ne '-' ) {
@@ -1088,7 +1125,7 @@ sub process_tc_filter() {
 		    fatal_error "ICMP not allowed with IPv6" unless $family == F_IPV4;
 		    fatal_error "SOURCE PORT(S) are not allowed with ICMP" if $sportlist ne '-';
 
-		    my ( $icmptype , $icmpcode ) = split '//', validate_icmp( $portrange );
+		    my ( $icmptype , $icmpcode ) = split '/', validate_icmp( $portrange );
 
 		    my $rule1 = "   match icmp type $icmptype 0xff";
 		    $rule1   .= "\\\n   match icmp code $icmpcode 0xff" if defined $icmpcode;
@@ -1099,7 +1136,7 @@ sub process_tc_filter() {
 		    fatal_error "IPv6 ICMP not allowed with IPv4" unless $family == F_IPV4;
 		    fatal_error "SOURCE PORT(S) are not allowed with IPv6 ICMP" if $sportlist ne '-';
 
-		    my ( $icmptype , $icmpcode ) = split '//', validate_icmp6( $portrange );
+		    my ( $icmptype , $icmpcode ) = split '/', validate_icmp6( $portrange );
 
 		    my $rule1 = "   match icmp6 type $icmptype 0xff";
 		    $rule1   .= "\\\n   match icmp6 code $icmpcode 0xff" if defined $icmpcode;
@@ -1181,6 +1218,8 @@ sub process_tcfilters() {
 
     my $fn = open_file 'tcfilters';
 
+    our $lastdevice = '';
+
     if ( $fn ) {
 	my @family = ( $family );
 	
@@ -1208,6 +1247,12 @@ sub process_tcfilters() {
 	}
 
 	Shorewall::IPAddrs::initialize( $family = pop @family );
+
+	if ( $lastdevice ) {
+	    pop_indent;
+	    emit "fi\n";
+	}
+
     }
 }
 
@@ -1285,7 +1330,7 @@ sub setup_simple_traffic_shaping() {
 	first_entry
 	    sub {
 		progress_message2 "$doing $fn1...";
-		warning_message "There are entries in $fn1 but $fn was empty" unless $interfaces;
+		warning_message "There are entries in $fn1 but $fn was empty" unless $interfaces || $family == F_IPV6;
 	    };
 
 	process_tc_priority while read_a_line;
@@ -1373,26 +1418,9 @@ sub setup_traffic_shaping() {
 		      qq(fi) );
 	    }
 
-	    my $in_burst = '10kb';
-	    my $inband;
-
-	    if ( $devref->{in_bandwidth} =~ /:/ ) {
-		my ( $in_band, $burst ) = split /:/, $devref->{in_bandwidth}, 2;
-
-		if ( defined $burst && $burst ne '' ) {
-		    fatal_error "Invalid IN-BANDWIDTH" if $burst =~ /:/;
-		    fatal_error "Invalid burst ($burst)" unless $burst =~ /^\d+(k|kb|m|mb|mbit|kbit|b)?$/;
-		    $in_burst = $burst;
-		}
-
-		$inband = rate_to_kbit( $in_band );
-	    } else {
-		$inband = rate_to_kbit $devref->{in_bandwidth};
-	    }
-
-	    if ( $inband ) {
+	    if ( $devref->{in_bandwidth} ) {
 		emit ( "run_tc qdisc add dev $device handle ffff: ingress",
-		       "run_tc filter add dev $device parent ffff: protocol all prio 10 u32 match ip src 0.0.0.0/0 police rate ${inband}kbit burst $in_burst drop flowid :1"
+		       "run_tc filter add dev $device parent ffff: protocol all prio 10 u32 match ip src 0.0.0.0/0 police rate $devref->{in_bandwidth}kbit burst $devref->{in_burst} drop flowid :1"
 		     );
 	    }
 
