@@ -26,7 +26,7 @@
 #       You may only use this script to uninstall the version
 #       shown below. Simply run this script to remove Shorewall Firewall
 
-VERSION=4.4.15.1
+VERSION=4.4.19.1
 
 usage() # $1 = exit status
 {
@@ -72,6 +72,9 @@ else
     VERSION=""
 fi
 
+[ -n "${LIBEXEC:=share}" ]
+[ -n "${PERLLIB:=share/shorewall}" ]
+
 echo "Uninstalling shorewall $VERSION"
 
 if qt iptables -L shorewall -n && [ ! -f /sbin/shorewall-lite ]; then
@@ -85,7 +88,9 @@ else
 fi
 
 if [ -n "$FIREWALL" ]; then
-    if [ -x /sbin/insserv -o -x /usr/sbin/insserv ]; then
+    if [ -x /usr/sbin/updaterc.d ]; then
+	updaterc.d shorewall remove
+    elif [ -x /sbin/insserv -o -x /usr/sbin/insserv ]; then
         insserv -r $FIREWALL
     elif [ -x /sbin/chkconfig -o -x /usr/sbin/chkconfig ]; then
 	chkconfig --del $(basename $FIREWALL)
@@ -104,6 +109,8 @@ rm -rf /etc/shorewall
 rm -rf /etc/shorewall-*.bkout
 rm -rf /var/lib/shorewall
 rm -rf /var/lib/shorewall-*.bkout
+rm -rf /usr/$PERLLIB}/Shorewall/*
+rm -rf /usr/${LIBEXEC}/shorewall
 rm -rf /usr/share/shorewall
 rm -rf /usr/share/shorewall-*.bkout
 rm -rf /usr/share/man/man5/shorewall*
